#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_TIME_H
#include <time.h>
#endif

#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef	HAVE_NETDB_H
#include <netdb.h>
#endif

#ifdef	HAVE_FCNTL_H
#include <fcntl.h>
#endif

#include <strfunc.h>

#include <desire.h>

char buf[128];

char ipfrom[15];
char ipto[15];
int pckt;
int bytes;

int readconfig();

#define	IFDRY	1
#define	STOP	2

typedef struct _nets {
	struct in_addr ip;
	struct in_addr mask;
	struct in_addr ipto;
	struct in_addr maskto;
	char *sys;
	unsigned int flags;
	unsigned int inb, outb;
	struct _nets *next;
} nets;

static	ctb_data	rec;

nets *net_head=NULL;
nets *net_tail=NULL;

int sux(char *);

int main(int ac, char **av) {
	int s;
	FILE *fd;
	int z, l;
	struct in_addr ip1, ip2;
	nets *nh;
	int avs;
	char *host;
	time_t tloc;
	time_t age = 0;
	int demo = 0;
	ctb_session *ctbs;

	if(ac < 2)
		sux("Usage: cistat <nas_hostname> [...]");

	if(getenv("DEMO")) {
		fprintf(stderr, "Demo mode. Accounting not cleared.\n");
		demo = 1;
	}

	if(!(ctbs=ctb_open(NULL, "traffic", NULL))) {
		fprintf(stderr, "Can't open DESIRE session.\n");
		exit(2);
	};
	readconfig();


	/*
	 * Loop and gather information from specified hosts.
	 */
	for(avs=1; avs < ac; avs++) {
		host=av[avs];

		time(&tloc);

		if(demo) {
			s=rcmd(&host, 514, "root", "root", "sh ip acco", NULL);
			if(s == -1) sux("Connection could not be established.");
			printf("Hostname: %s\n", host);
		} else {
			s=rcmd(&host, 514, "root", "root", "cle ip acco", NULL);
			if(s == -1) sux("Connection could not be established.");
			while(read(s, buf, sizeof(buf)) > 0);
			close(s);
	
			s=rcmd(&host, 514, "root", "root", "sh ip acco che", NULL);
			if(s == -1) sux("Connection could not be established.");
			printf("Hostname: %s\n", host);
		}
	
		fd=fdopen(s, "r+");
		if(!fd) {
			close(s);
			sux("Can't fdopen().");
		}

#define ADA     "Accounting data age is "
#define ADA_SIZE        sizeof(ADA)-1
#define ADAE     "Accounting data age exact "	/* IPCAD extension */
#define ADAE_SIZE        sizeof(ADAE)-1

		while(fgets(buf, sizeof(buf), fd)) {
	
			if(*buf != ' ') {
				if(strncmp(buf, ADA, ADA_SIZE) == 0) {
					age = 60 * strtol(buf + ADA_SIZE, NULL, 10);
					continue;
				};
				if(strncmp(buf, ADAE, ADAE_SIZE) == 0) {
					age = strtol(buf + ADAE_SIZE, NULL, 10);
					continue;
				};
				continue;
			};
	
			if(sscanf(buf, "%s %s %d %d\n", ipfrom, ipto, &pckt, &bytes)!=4)
				continue;
	
			if(!inet_aton(ipfrom, &ip1) || !inet_aton(ipto, &ip2))
				continue;
	
			printf("%s\t-> ", inet_ntoa(ip1));
			printf("%-15s: %-8d ( ", inet_ntoa(ip2), bytes);
	
			for(z=0, nh=net_head; nh; nh=nh->next) {
				l=0;
				if(z && (nh->flags & IFDRY))
					continue;
				if((nh->ip.s_addr == (ip1.s_addr & nh->mask.s_addr)) &&
				nh->ipto.s_addr == (ip2.s_addr & nh->maskto.s_addr)) {
					nh->outb+=bytes;
					l=z=1;
				};
				if((nh->ip.s_addr == (ip2.s_addr & nh->mask.s_addr)) &&
				nh->ipto.s_addr == (ip1.s_addr & nh->maskto.s_addr)) {
					nh->inb+=bytes;
					l=z=1;
				};
				if(l)
					printf("%s ", nh->sys);
	
				if(l && (nh->flags & STOP))
					break;
			};
	
			printf(")\n");
		};
	
		fclose(fd);
		printf("\n");
	

		/*
		 * Saving data.
		 */
		for(nh=net_head; nh; nh=nh->next) {
			if(nh->inb || nh->outb) {
				printf("%s\t<->\t", inet_ntoa(nh->ip));
				printf("%s\t(%s)\t%d/%d\n",
				    inet_ntoa(nh->ipto), nh->sys, nh->inb, nh->outb);
			} else {
				continue;
			};
	
			if(ctb_chname(ctbs, nh->sys)) {
				fprintf(stderr,
					"Can't choose DESIRE name: cisco/%s. Ignored.\n",
					nh->sys);
				continue;
			};

			/* Record stats */
			rec.TS=tloc;
			rec.In=nh->inb;
			rec.Out=nh->outb;
			rec.U1=age;

			if(!demo)
				if(ctb_add(ctbs, &rec))
					fprintf(stderr, "Can't add DESIRE data.\n");

			/* Release locks */
			ctb_chname(ctbs, NULL);

			/* Clear stats */
			nh->inb = 0;
			nh->outb = 0;
		};

	};	/* for(hosts) */

	ctb_close(ctbs);

	return 0;
};


int
readconfig() {
	FILE *cf;
	char *p;
	struct in_addr ip, ipto;
	struct in_addr mask, maskto;
	nets *nh;
	sbuf *sb;
	slist *sl;

	if(!(cf=fopen("/usr/local/etc/cistat.conf", "r")))
		sux("Can't read config.");

	sb = sbuf_init();
	sl = sinit();

#define	C_SIP1	(sl->list[0])
#define	C_SIP2	(sl->list[1])
#define	C_SYS	(sl->list[2])

	while((sbuf_zero(sb) == 0) && sbuf_fgets(sb, cf)) {
		if((p=strchr(sb->buf, '#'))) *p=0;
		sclear(sl);
		splitf(sl, sb->buf, " \t\n", 0);
		if(sl->count < 3)
			continue;

		if(strchr(C_SYS, '/'))
			sux("Slashes in system name.");

		if( split_network(C_SIP1,
			(unsigned int *)&ip.s_addr, (unsigned int *)&mask.s_addr) ) {
			sux("Invalid ip/mask.");
		};

		if( split_network(C_SIP2,
			(unsigned int *)&ipto.s_addr, (unsigned int *)&maskto.s_addr) ) {
			sux("Invalid ip/mask.");
		};

		printf("%s/", inet_ntoa(ip));
		printf("%s\t<-> ", inet_ntoa(mask));
		printf("%s/", inet_ntoa(ipto));
		printf("%s\t-> %s", inet_ntoa(maskto), C_SYS);

		nh = (nets *)sf_calloc(1, sizeof(nets));
		if(net_tail) {
			net_tail = net_tail->next = nh;
		} else {
			net_head = net_tail = nh;
		}

		nh->ip=ip;
		nh->mask=mask;
		nh->ipto=ipto;
		nh->maskto=maskto;
		nh->sys=strdup(C_SYS);
		if(!nh->sys) {
			perror("strdup()");
			exit(2);
		}
		nh->next=NULL;

		if(sl->count > 3) {
			int i;
			printf("\t(Flags: ");
			for(i = 3; i < sl->count; i++) {
				if(i != 3) printf(", ");
				if(!strcasecmp(sl->list[i], "ifdry")) {
					printf("%s", "IF_DRY");
					nh->flags |= IFDRY;
				} else if(!strcasecmp(sl->list[i], "stop")) {
					printf("%s", "STOP");
					nh->flags |= STOP;
				} else {
					fprintf(stderr, "%s: ", sl->list[i]);
					sux("unknown token.");
				};
			};
			printf(")");
		};


		printf("\n");

	};

	printf("\n");

	sbuf_free(sb);

	if(!net_head)
		sux("Empty config.");

	fclose(cf);

	return 0;
};


int
sux(char *s) {
	fprintf(stderr, "%s\n", s);
	exit(2);
};
