/* $Id: mg_parser.c,v 1.2 2001/04/10 13:49:38 vlm Exp $ */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sysexits.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <strfunc.h>
#include <desire.h>

#define	FROM  0
#define	TO    1


extern slist *localdomains;
extern ctb_session *ctbs;

unsigned int map_mta(char *);
unsigned int map_user(char *);
unsigned int map_domain(char *);
unsigned int make_hash(char *);

struct _mail_descriptor {
	char *mta;	/* MTA serving message */
	char *email;	/* Email address that we work on */

	time_t time;	/* Time when done */

	int fromto;	/* from = 0, to = 1 */
	size_t nrcpts;	/* Number of recipients if fromto = 0 */

	unsigned int relay_ip;	/* 0 means local mailer */

	size_t size;	/* Message size */

	char queueid[32];
};

int
test_ldomain(char *email, char **domain) {
	char *p;

	if(!email || !domain)
		return 0;

	p = strchr(email, '@');
	if(!p) {
		return 1;
	}

	*p++ = 0;

	if(!*p) {
		*domain = NULL;
		return 1;
	}

	if(!localdomains || localdomains->count == 0)
		return 0;

	*domain = p;

	if(scfind(localdomains, p) != -1)
		return 1;

	return 0;
}


/* return a@d from <a@d> */
char *
strip_angles(char *str) {
	char *p;
	if(!(p = strchr(str, '<')))
		return str;

	str = p + 1;

	for(p = str; *p; p++) {
		if(*p == '>')
			break;
		*p = tolower(*p);
	};
	if(!*p)
		return NULL;

	*p = 0;

	return str;
}

char *
strip_relay(char *str) {
	char *p;
	if(!(p=strchr(str, '[')))
		return NULL;

	str = p + 1;

	if(!*str || !(p=strchr(str, ']')))
		return NULL;

	*p = 0;

	return str;
}

ctb_data *
parser(char *str) {
	static ctb_data data;
	static slist *sl = NULL;
	static sed_t *se = NULL;
	slist *sr;
	time_t tme;
	char *mta;
	char *queueid;
	int fromto = -1;
	char *email;
	char **vp;
	unsigned int relay_ip = 0;
	size_t nrcpts = 0;
	size_t size = 0;
	char *domain = NULL;

	if(sl) sclear(sl);
	else sl = sinit();

	if(!se) {
		se = sed_compile("/^([a-z]+) ([^:]+): ([^:]+): (.*)$/rein");
		if(!se) {
			fprintf(stderr, "Can't compile regular expression.\n");
			exit(0);
		}
	}

	/* Main processing */

	if(!(tme = strtotime(str)))
		return NULL;

	str += 16;

	if(!sed_exec(se, str))
		return NULL;

	sr = sed_results(se);
	if(!sr)
		return NULL;

	mta = sr->list[0];
	queueid = sr->list[2];

	/* Splitting parameters by ", " */
	splitf(sl, sr->list[3], ", ", 2);
	if(sl->count < 2)
		return NULL;

	email = sl->list[0];

	if(strncmp(email, "to=", 3) == 0) {
		fromto = 1;
		email += 3;
	} else if(strncmp(email, "from=", 5) == 0) {
		fromto = 0;
		email += 5;
	} else {
		return NULL;
	}

	if(!(email = strip_angles(email)) || !*email)
		return NULL;

	for(vp = &sl->list[1]; *vp; vp++) {
		if(strncmp(*vp, "size=", 5) == 0) {
			errno = 0;
			size = atoi(*vp + 5);
			/* We don't know size for sure, so don't guess */
			if(errno) size = 0;
		} else if(strncmp(*vp, "nrcpts=", 7) == 0) {
			errno = 0;
			nrcpts = atoi(*vp + 7);
			if(errno) nrcpts = 0;
		} else if(strncmp(*vp, "nrcpt=", 6) == 0) {
			errno = 0;
			nrcpts = atoi(*vp + 6);
			if(errno) nrcpts = 0;
		} else if(strncmp(*vp, "relay=", 6) == 0) {
			char *relay;
			relay = strip_relay(*vp + 6);
			if(!relay)
				continue;
			inet_aton(relay, (struct in_addr *)&relay_ip);
		} else if(fromto && (strncmp(*vp, "stat=", 5) == 0)) {
			if(strcasecmp(*vp, "stat=Sent"))
				return NULL;
		}

		if(relay_ip && size && (fromto || nrcpts)) break;
	};

	if(!fromto && !nrcpts)
		return NULL;

	memset(&data, 0, sizeof(ctb_data));

	data.TS = tme;

	data.FL.c.s1c1 = fromto;
	data.FL.c.s1c2 = map_mta(mta);
	data.FL.s.s2 = nrcpts;

	if(test_ldomain(email, &domain))
		data.In = map_user(email);

	data.Out = map_domain(domain);

	data.Src = relay_ip;
	data.Dst = 0 /* Unused */;

	data.U1 = make_hash(queueid);
	data.U2 = size;

	return &data;
};


ctb_map_cache *cache_mta = NULL;
ctb_map_cache *cache_user = NULL;
ctb_map_cache *cache_domain = NULL;



unsigned int
map_user(char *user) {
	int id;

	if(!user || !*user)
		return 0;

	id = ctb_map_cache_get(&cache_user, user);
	if(id) return id;

	id = ctb_revmap(ctbs, FIELD_IN, user, 0, NULL, NULL);

	if(!id)
		id = ctb_addmap(ctbs, FIELD_IN, 0, user);

	if(id) ctb_map_cache_put(&cache_user, user, id, 100);

	return id;
}

unsigned int
map_domain(char *domain) {
	int id;

	if(!domain || !*domain)
		return 0;

	id = ctb_map_cache_get(&cache_domain, domain);
	if(id) return id;

	id = ctb_revmap(ctbs, FIELD_OUT, domain, 0, NULL, NULL);

	if(!id)
		id = ctb_addmap(ctbs, FIELD_OUT, 0, domain);

	if(id) ctb_map_cache_put(&cache_domain, domain, id, 100);

	return id;
}


unsigned int
map_mta(char *mta) {
	int id;

	if(!mta || !*mta)
		return 0;

	id = ctb_map_cache_get(&cache_mta, mta);
	if(id) return id;

	id = ctb_revmap(ctbs, 5, mta, 0, NULL, NULL);

	if(!id)
		id = ctb_addmap(ctbs, 5, 0, mta);

	if(id) ctb_map_cache_put(&cache_mta, mta, id, 100);

	return id;
}


/*
 * Phong's linear congruential hash
 */
#define dcharhash(h, c) ((h) = 0x63c63cd9*(h) + 0x9c39c33d + (c))

unsigned int
make_hash(char *str) {
	size_t len;
	register const u_char *e, *key;
	register u_int32_t h;
	register u_char c;

	if(!str)
		return 0;

	key = str;
	len = strlen(key);
	e = key + len;
	for (h = 0; key != e;) {
		c = *key++;
		if (!c && key > e)
			break;
		dcharhash(h, c);
	}

	return h;
}

