#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <errno.h>
#include <syslog.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef	HAVE_SYS_FILE_H
#include <sys/file.h>
#endif

#ifdef HAVE_LIBCOST
#include <cost.h>

int _lc_cost_initialized = 0;
char **_lc_cost_cls = NULL;
double dataCost = 0.0;
#endif

#ifdef	HAVE_GD_H
#include <gd.h>
#endif

#ifdef	HAVE_GDFONTS_H
#include <gdfonts.h>
#endif

#ifdef	HAVE_GDFONTT_H
#include <gdfontt.h>
#endif

#include <strfunc.h>

#include "desire.h"

#include "ddraw.h"


static char mime_types[256];
int strict_mime_types = 0;

static char domain_hashes[256];
unsigned int *domains = NULL;
size_t domains_count = 0;
int domains_error = 0;

int add_domain_index(unsigned int, char const *, void *);

/* Offset, Mask, Value, Rule, Function */
struct line_filter filter_direct = { FIELD_FL, 0x000000ff, 1, 1, NULL };
struct line_filter filter_cache  = { FIELD_FL, 0x000000ff, 0, 1, NULL };
struct line_filter filter_proxy  = { FIELD_FL, 0x000000fe, 0, 0, NULL };
struct line_filter filter_except = { FIELD_FL, 0x00ff0000, 0, 0, NULL };

struct custom_line lines[] = {
	/* All */
  { 0x606060, 0xf0f0f0, FIELD_IN, -1, FIELD_OUT, -1, -1, NULL, 0 },
	/* Direct */
  { 0x605010, -1, FIELD_IN, -1, FIELD_OUT, -1, -1, &filter_direct, 0 },
	/* Proxy */
  { 0x28a000, -1, FIELD_IN, -1, FIELD_OUT, -1, -1, &filter_proxy, 0 },
	/* Cache */
  { 0x0032ff, -1, FIELD_IN, -1, FIELD_OUT, -1, -1, &filter_cache, 0 },
	/* eXceptions */
  { 0xff3200, -1, FIELD_IN, -1, FIELD_OUT, -1, -1, &filter_except, 0 },
};

void preFilter(ctb_session *, ctb_data *, int, struct desire_image_options *);

char buf[1024];

int dc;
ctb_data *dt;
int mix;
unsigned int ip;
unsigned int mask;
unsigned int *domain;
unsigned int dmn;
int dmc;

int
main(int ac, char **av) {
	struct desire_image_options *io;
	ctb_dss *dss = NULL;
	gdImagePtr im;
	ctb_session *ctbs = NULL;
	int lockMyself;
	slist *sl;
	float fl;
	int i;

	printf("Expires: %s\n", timetostr(time(NULL)-300,
	TFMT_RFC822 | TFMT_OLD822GMT));
#ifdef	HAVE_GDIMAGEPNG
	printf("Content-type: image/png\n\n");
#else
	printf("Content-type: image/gif\n\n");
#endif
	fflush(stdout);

	if( (lockMyself = open(av[0], O_RDONLY)) != -1)
		flock(lockMyself, LOCK_EX);

	io=fill_desire_image_options(0);
	if(!io->comments)
		io->comments = sinit();
	io->legend_factor = 1024;
	io->vertical_legend = sf_strdup("KBytes per second");

	if(io->mix1 >= 0) {
		filter_except.Mask =
			(filter_except.Mask & 0xff00ffff)
			| ((io->mix1 & 0xff) << 16);
	};

	if((ctbs=ctb_open(NULL, "proxy", io->file))) {
		dss = ctb_get_dss(ctbs, &io->range);

		if(io->list1) {
			for(i = 0; i < io->list1->count; i++)
				ctb_revmap(ctbs, FIELD_DST, io->list1->list[i], RVM_FUZZY, add_domain_index, io);
			sprintf(buf, "%s (%d %s)",
				sjoin(io->list1, ", "),
				domains_count,
				(domains_count==1)?"domain":"domains"
			);
			sadd(io->comments, buf);
		};

		if((sl=param2("mime", 1))) {
			unsigned int t;
			int k;
			for(i = 0; i < sl->count; i++) {
				errno = 0;
				t = strtoul(sl->list[i], NULL, 0);
				if(errno || t > 255) continue;
				mime_types[t] = 1;
				strict_mime_types = 1;
				if(!(t & 0x0f))
					for(k = 0; k <= 0x0f; k++)
						mime_types[(t & 0xf0) + k] = 1;
			};
		};

	};

	mix = io->mix2;
	ip = io->ip;
	mask = io->mask;

#ifdef HAVE_LIBCOST
if(io->costtype && (lc_init(NULL) == 0)) {

	_lc_cost_initialized = 1;

	if(io->costclass)
		_lc_cost_cls=charize(io->costclass);

	sadd(io->comments, "Cost template");
};
#endif

	sadd(io->comments, "Proxy requests made: 0");


	/* Filter domains */
	io->optBlockPreFilter = preFilter;

	im=ddraw_dss(ctbs, dss, lines,
		sizeof(lines)/sizeof(struct custom_line), io);

	sprintf(buf, "Display time: %ld.%03ld",
		io->elapsed.tv_sec, io->elapsed.tv_usec/1000);
	gdImageString(im, gdFontTiny, io->image_width - 5 * (strlen(buf) + 2) , io->image_height - 24 , buf, 0);


	/* All */
	sprintf(buf, "A: %quk", lines[0].all >> 10);
		gdImageString(im, gdFontSmall, 6, 95, buf, lines[0].foreColor);

	fl = lines[0].meaningful_points
		?(float)(lines[0].all >> 10)/lines[0].meaningful_points/io->inpixel:0;
	sprintf(buf, "   %0.2fk/s", fl);
		gdImageString(im, gdFontSmall, 6, 105, buf, lines[0].foreColor);

	fl = lines[0].all?((float)lines[1].all * 100 / lines[0].all):0;
	sprintf(buf, "D: %.02f%%", fl);
		gdImageString(im, gdFontSmall, 6, 115, buf, lines[1].foreColor);
	fl = lines[0].all?((float)lines[2].all * 100 / lines[0].all):0;
	sprintf(buf, "P: %.02f%%", fl);
		gdImageString(im, gdFontSmall, 6, 125, buf, lines[2].foreColor);
	fl = lines[0].all?((float)lines[3].all * 100 / lines[0].all):0;
	sprintf(buf, "C: %.02f%%", fl);
		gdImageString(im, gdFontSmall, 6, 135, buf, lines[3].foreColor);
	fl = lines[0].all?((float)lines[4].all * 100 / lines[0].all):0;
	sprintf(buf, "X: %.02f%%", fl);
		gdImageString(im, gdFontSmall, 6, 145, buf, lines[4].foreColor);


	if(isatty(1))
		fprintf(stderr, "[Image will be displayed here]\n");
	else {
#ifdef	HAVE_GDIMAGEPNG
		gdImagePng(im, stdout);
#else
		gdImageGif(im, stdout);
#endif
	}

	gdImageDestroy(im);

	return 0;
};


int
add_domain_index(unsigned int Idx, char const *Value, void *optKey) {
	struct desire_image_options *io = (struct desire_image_options*)optKey;
	int *d;

	if(!io) return -1;

	d = realloc(domains, sizeof(*domains) * (++domains_count));
	if(!d) {
		domains_error = 1;
		domains_count = 0;
		return -1;
	};
	domains = d;

	domain_hashes[(unsigned char)(Idx >> 24)] = 1;

	domains[domains_count - 1] = Idx;

	return 0;
};


void
preFilter(ctb_session *ctbs, ctb_data *data, int dataCount, struct desire_image_options *io) {

	if(!data) {
		int i;

		if((i = scfind(io->comments, "Proxy requests made: 0")) != -1) {
			sdel(io->comments, i);
			if(!lines[0].elements)
				sprintf(buf, "No proxy requests found");
			else if(lines[0].elements == lines[0].elements_filtered)
				sprintf(buf, "%d proxy request%s made",
					lines[0].elements, (lines[0].elements>1)?"s":"");
			else
				sprintf(buf, "%d request%s selected from %d proxy request%s",
				lines[0].elements_filtered, lines[0].elements_filtered>1?"s":"",
				lines[0].elements, lines[0].elements>1?"s":"");
			sins(io->comments, buf, i);
		}

#ifdef HAVE_LIBCOST
		if(_lc_cost_initialized &&
			(i = scfind(io->comments, "Cost template")) != -1) {
			sdel(io->comments, i);
			sprintf(buf, "Traffic cost: %.2f$", dataCost);
			sins(io->comments, buf, io->comments->count - 1);
		}
#endif
	}

	for(dc = dataCount, dt = data; dc; dt++, dc--) {

		dt->Out = dt->Out / 1000;	/* Msecs to seconds */

		if(mask) {
			if((dt->Src & mask) != ip)  {
				dt->TS = 0;
				dt->In = 0;
				continue;
			};
		};

		if(mix > 0) {
			switch(dt->FL.c.s2c2 & 0x03) {
			case 0x01: if(!(mix & 1)) dt->TS = 0; break; /* HTTP */
			case 0x02: if(!(mix & 2)) dt->TS = 0; break; /* FTP */
			case 0x03: if(!(mix & 4)) dt->TS = 0; break; /* CONNECT */
			case 0x00: if(!(mix & 8)) dt->TS = 0; break; /* UNKNOWN */
			};
			if(!dt->TS) {
				dt->In = 0;
				continue;
			};
		};

		if(strict_mime_types) {
			if(!mime_types[(unsigned char)dt->FL.c.s1c2]) {
				dt->TS = 0;
				dt->In = 0;
				continue;
			};
		};

		/* Check for specified domains */
		if(domains_count) {
			dmn = dt->Dst;

			if(!domain_hashes[(unsigned char)(dmn >> 24)]) {
				dt->TS = 0;
				dt->In = 0;
				continue;
			};

			for(domain=domains,
				dmc=domains_count;
					dmc; dmc--, domain++) {
				if(*domain != dmn)
					continue;

				if((domain - domains) < 8)
					break;

				/* Shift data */
				for(; domain > domains; domain--)
					*domain = domain[-1];

				*domains = dmn;
				dmc = 0;
				break;
			};

			if(!dmc) {
				dt->TS = 0;
				dt->In = 0;
				continue;
			};
		};

	};


#ifdef HAVE_LIBCOST
	if(_lc_cost_initialized)
		dataCost += lc_cost(data, dataCount, sizeof(ctb_data),
			FIELD_TS, FIELD_IN, sizeof(int),
				0, io->costtype, _lc_cost_cls);
#endif

};
