/*
 * Cool Technical Desire, general accounting and statistics API.
 *
 * Copyright (c) 2000, 2001 Lev Walkin, Alexandr Grigoryev, Vlad Skvortsov.
 *
 */

#ifndef	__DESIRE_H__
#define	__DESIRE_H__

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>

#include "ctbdefs.h"

/******************************************/
/* Initialize and open accounting session */
/******************************************/
typedef void ctb_session;

ctb_session *
ctb_open(
	const char *optConfigName,	/* /usr/local/etc/desire.conf */
	const char *Service,	/* Service name */
	const char *optName	/* Optional subdestination name */
);

/* Change destination name (optName from ctb_open) within the service */
int	ctb_chname(ctb_session *, const char *name);
void	ctb_close(ctb_session *);

/* Get CTB session description */
char *ctb_description(ctb_session *);


/*
** Add this structure to the CTB file
** Note: it locks the last used file descriptor!
** Be sure to ctb_close() or ctb_chname() when you're finished!
*/
int	ctb_add(ctb_session *, void *data);

/* Synchronize with on-disk state */
void ctb_sync(ctb_session *ctbs);


/* Get the specified data range of data */
void *ctb_get(ctb_session *, ctb_range *range, int *count);

/* Initialize Data Sources Structure */
ctb_dss *ctb_get_dss(ctb_session *, ctb_range *range);

/*
 * Get another size bytes from the Data Sources.
 * Return number of entries, or <= 0 when data sources are empty.
 */
ssize_t ctb_get_next(ctb_session *, ctb_dss *dss, void *data, size_t size);

/*
 * Always call ctb_free_dss() when finished working with ctb_get_next().
 */
void ctb_free_dss(ctb_dss *);


/********************/
/* Formatted output */
/********************/

char *ctb_format(ctb_session *, const char *format, void *data);


/*****************/
/*** BYTE MAPS ***/
/*****************/

/*
** If string pointer is a NULL value, than map
** will be deleted instead of replacing.
** ctb_addmap() will return 0 in case of any error, or WandIndex.
*/
int
ctb_addmap(
	ctb_session *,
	int FIELD,
		/* Desired index to be assigned to string */
	unsigned int optWantIndex,
	const char *string
);

/* Map binary value to string */
char	*ctb_map(ctb_session *, void *data, int FIELD);

/* rvmodeFlag */
#define	RVM_EXACT	0x00
#define	RVM_FUZZY	0x01	/* Use fnmatch() on value */

/* Map string to binary value */
unsigned int
ctb_revmap(
	ctb_session *,
	int FIELD,
	const char *value,
	int rvmodeFlag,		/* See RVM_* flags in structs.h */
	int (*opt_fuzzy_return)(
		unsigned int Idx,
		char const *resolvedValue,
		void *optRecvKey
	),
	void *optSendKey
);


/**************/
/* MAPS CACHE */
/**************/

/* ctb_map_cache structure defined in structs.h */

unsigned int ctb_map_cache_get(
	ctb_map_cache **cache,
	char *value
);

unsigned int ctb_map_cache_put(
	ctb_map_cache **cache,
	char *value,
	unsigned int idx,
	int maxgrow	/* Maximum cache size */
);


/********************/
/* Helper functions */
/********************/

/* Get entry chunk size for the specified session */
unsigned int ctb_entry_size(ctb_session *);

/* Get the last used number */
unsigned int ctb_map_lastno(ctb_session *, int FIELD);

/* Get the number of the entries in map */
ssize_t ctb_map_size(ctb_session *, int FIELD);


/*
** Fills specified time range with appropriate values
** according to CLP_HDWMY. See above.
*/
ctb_range *ctb_fill_time(ctb_range *optRange, int CLP_HDWMY);


/*************************/
/*** Files multiplexor ***/
/*************************/

/*
** Read n files and multiplex their contents into one Desire session
** Note: Not all the contents of all specified files may be fully read.
** optPosExt specify the extensions to be used by position cache file.
*/
int ctb_mux(
	ctb_session *ctbs,
	ctb_data *(*ParserFunc)(char *StringToParse),
	int Mux_Flags,	/* See below (CTB_MUX_*) */
	char *fnames[],
	char *optPosExt
);

/* Ignore errors on file opens */
#define	CTB_MUX_IGN_FOF   0x01

/* Truncate file when EOF reached */
#define	CTB_MUX_TRUNC     0x02

/* Do not wait for other files that are ended (Dangerous) */
#define	CTB_MUX_NOWAIT    0x04

/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

/* Internal functions */

off_t	ctb_findstart (FILE *fp, time_t from, off_t fsize, int EntrySize, time_t *tstamp);
off_t	ctb_findend   (FILE *fp, time_t till, off_t fsize, int EntrySize, time_t *tstamp);

#endif
