#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "desire.h"
#include "session.h"

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef	HAVE_INTTYPES_H
#include <inttypes.h>
#endif

#ifdef	HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif

#define hashf(h,c) ((h) = 0x63c63cd9*(h) + 0x9c39c33d + (c))


unsigned int
_ctb_make_hash(char *key) {
	register unsigned int hash;
	register unsigned char c;

	for(hash = 0; (c=*key); key++)
		hashf(hash, c);

	return hash;
}


unsigned int
ctb_map_cache_get(ctb_map_cache **cachep, char *string) {
	unsigned int hash;
	ctb_map_cache *cache;
	ctb_map_cache *cprev;

	if(!string || !cachep || !(cache = *cachep))
		return 0;

	hash = _ctb_make_hash(string);

	/* Try to find in cache */
	for(cprev = NULL; cache;
			cprev = cache, cache = cache->next) {
		if(
			(cache->ValueHash != hash)
			|| strcasecmp(cache->Value, string)
		) continue;

		/* Move this entry to the top */
		if(cprev) {
			/* Collapse the previous and the next */
			cprev->next = cache->next;
			cache->next = *cachep;	/* One up */
			*cachep = cache;
		}
		return cache->Index;
	}

	return 0;
}


unsigned int
ctb_map_cache_put(ctb_map_cache **cachep, char *string, unsigned int idx, int maxgrow) {
	ctb_map_cache *cprev;
	ctb_map_cache *cache;

	if(!string || (maxgrow < 1) || !cachep)
		return -1;

	cache = (ctb_map_cache *)malloc(sizeof(ctb_map_cache));
	if(!cache)
		return -1;

	cache->Value = strdup(string);
	if(!cache->Value) {
		free(cache);
		return -1;
	}

	cache->ValueHash = _ctb_make_hash(string);
	cache->Index = idx;
	cache->next = *cachep;
	*cachep = cache;

	for(cprev = NULL; maxgrow >= 0 && cache;
		cprev = cache, cache = cache->next, maxgrow--);

	if(!cprev || !cache)
		return 0;

	/* Crop tail */
	cprev->next = NULL;

	/* Delete tails */
	while(cache) {
		cprev = cache->next;
		free(cache->Value);
		free(cache);
		cache = cprev;
	};

	return 0;
}


