#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>

#include "desire.h"
#include "ctbint.h"
#include "session.h"

#ifdef	CTB_RETURN
#undef	CTB_RETURN
#endif

#define	CTB_RETURN { \
	free(fname); \
	fchdir(rememberdirfd); \
	close(rememberdirfd); \
	return NULL; }

int _ctb_add_ds(ctb_dss *, int fdc, off_t start, off_t end);

ctb_dss *
ctb_get_dss(ctb_session *ctbsp, ctb_range *range) {
	int  rememberdirfd;	/* Remember current position */
	ctb_sess *ctbs = (ctb_sess *)ctbsp;
	ctb_dss *dss = NULL;
	char *fname;
	FILE *fp;
	struct tm *tm;
	struct stat sb;
	int starty, startm, endy, endm;
	int cy, cm, cme;
	off_t startof, endof;

	/* Session not prepared yet */
	if(!ctbs || !range)
		return NULL;

	/* Time range checks */
	if(range->from > range->to) {
		time_t tmp;
		tmp = range->to;
		range->to = range->from;
		range->from = tmp;
	};

	if(range->either == 1)
		if(!ctb_fill_time(range, range->simple_def))
			return NULL;


	/* Remember current folder */
	rememberdirfd=open(".", O_RDONLY);
	if(rememberdirfd == -1)
		return NULL;

	fname=(char *)malloc(strlen(ctbs->folder) + strlen(ctbs->name) + 32);
	if(!fname)	/* ENOMEM? */
		return NULL;

	/* Change folder to destination folder */
	sprintf(fname, "%s/%s", ctbs->folder, ctbs->name);
	if(chdir(fname))
		CTB_RETURN;

/****************************************************************/
/*** date-independent procedure of gathering file descriptors ***/
/****************************************************************/

	/* Initialize data sources structure */
	dss = (ctb_dss *)_ctb_calloc(1, sizeof(ctb_dss));
	if(!dss)
		CTB_RETURN;

	/* Start point */
	tm=localtime((time_t *)&(range->from));
	starty=tm->tm_year+1900;
	startm=tm->tm_mon+1;

	/* End point */
	tm=localtime((time_t *)&(range->to));
	endy=tm->tm_year+1900;
	endm=tm->tm_mon+1;

	/* Iterate years */
	for(cy = starty; cy <= endy; cy++) {

		/* Work with specified year */
		sprintf(fname, "%04d", cy);

		if((fp=fopen(fname, "r"))) {

			/* Strange things happen... */
			if(fstat(fileno(fp), &sb)) {
				fclose(fp);
				continue;
			}

			/* Defaulted to entire file */
			startof = 0;
			endof = sb.st_size;

			if(cy == starty)
				startof = ctb_findstart(fp, range->from,
					sb.st_size, ctbs->EntrySize, &dss->t_first);
			if(cy == endy)
				endof = ctb_findend(fp, range->to,
					sb.st_size, ctbs->EntrySize, &dss->t_last);

			if(_ctb_add_ds(dss, fileno(fp), startof, endof)) {
				ctb_free_dss(dss);
				fclose(fp);
				CTB_RETURN;
			};

			fclose(fp);
			continue;
		};

		if(cy == endy)
			cme=endm;
		else
			cme=12;

		for(cm=(cy==starty)?startm:1; cm <= cme; cm++) {

			sprintf(fname, "%04d-%02d", cy, cm);

			if(!(fp=fopen(fname, "r")))
				continue;

			if(fstat(fileno(fp), &sb)) {
				fclose(fp);
				continue;
			}

			startof = 0;
			endof = sb.st_size;

			if((cy == starty && cm == startm) || !dss->t_first)
				startof = ctb_findstart(fp, range->from,
					sb.st_size, ctbs->EntrySize, &dss->t_first);
			if((cy == endy && cm == endm) || !dss->t_last)
				endof = ctb_findend(fp, range->to,
					sb.st_size, ctbs->EntrySize, &dss->t_last);

			if(_ctb_add_ds(dss, fileno(fp), startof, endof)) {
				ctb_free_dss(dss);
				fclose(fp);
				CTB_RETURN;
			};

			fclose(fp);
		}; /* for(month) */
	}; /* for(years) */

	if(!dss->t_first)
		dss->t_first = range->from;

	if(!dss->t_last)
		dss->t_last = range->to;

	fchdir(rememberdirfd);
	close(rememberdirfd);

	return dss;
};


/**********************/
/* Service procedures */
/**********************/

off_t
ctb_findstart(FILE *fp, time_t from, off_t fsize, int EntrySize, time_t *tstamp) {
	time_t start = 0, end = 0, curt;
	off_t startof, endof, curof=-1, co;

	startof = 0;
	endof = fsize - (fsize % EntrySize) - EntrySize;

	/* fill start and end */
	fseek(fp, 0, SEEK_SET);
	fread(&start, sizeof(time_t), 1, fp);
	fseek(fp, endof, SEEK_SET);
	fread(&end, sizeof(time_t), 1, fp);

	if(from <= start) {
		if(tstamp)
			*tstamp = start;
		return 0;
	}
	if(from > end) {
		if(tstamp)
			*tstamp = 0;	/* There is no information */
		return fsize;
	}

  while(1) {

	co = curof;
	curof = startof + ((endof-startof) * (from - start)) / (end-start);
	curof -= curof % EntrySize;

	/* Pointed to the same location */
	if(co == curof) {
		while(1) {
			fseek(fp, curof, SEEK_SET);
			if(fread(&curt, sizeof(time_t), 1, fp) < 1)
				break;
			if(curt >= from)
				break;
			curof += EntrySize;
		};
		break;	/* return curof; */
	} /* if(co == curof) */

	fseek(fp, curof, SEEK_SET);
	fread(&curt, sizeof(time_t), 1, fp);

	if(curt == from)
		break;	/* return curof; */

	if(curt < from) { startof=curof; start=curt; }
	else { endof=curof; end=curt; };

	if(endof - startof == EntrySize) {
		if(tstamp)
			*tstamp = curt;
		curof = endof;
		break;
	}

  } /* while() */

	
	if(tstamp)
		*tstamp = curt;

	return curof;
};

off_t
ctb_findend(FILE *fp, time_t till, off_t fsize, int EntrySize, time_t *tstamp) {
	time_t start, end, curt;
	off_t startof, endof, curof=-1, co;

	startof = 0;
	endof = fsize - (fsize % EntrySize) - EntrySize;

	/* fill start and end */
	fseek(fp, 0, SEEK_SET);
	fread(&start, sizeof(time_t), 1, fp);
	fseek(fp, endof, SEEK_SET);
	fread(&end, sizeof(time_t), 1, fp);

	if(till <= start) {
		if(tstamp)
			*tstamp = 0;
		return 0;
	}
	if(till > end) {
		if(tstamp)
			*tstamp = end;
		return fsize;
	}

  while(1) {

	co=curof;
	curof = startof + ((endof-startof) * (till - start)) / (end-start);
	curof -= curof % EntrySize;

	if(co == curof) {
		while(1) {
			fseek(fp, curof, SEEK_SET);
			if(fread(&curt, sizeof(time_t), 1, fp) < 1)
				break;
			if(curt > till) {
				curt = till;
				break;
			}
			curof += EntrySize;
		};
		break;
	} /* if(co == curof) */
		
	fseek(fp, curof, SEEK_SET);
	fread(&curt, sizeof(time_t), 1, fp);

	if(curt == till) {
		curof += EntrySize;
		break;
	}
	if(curt < till) { startof=curof; start=curt; }
	else { endof=curof; end=curt; };

	if(endof - startof == EntrySize) {
		curof = endof;
		break;
	}

  } /* while() */

	if(tstamp)
		*tstamp = curt;

	return curof;
};

int
_ctb_add_ds(ctb_dss *dss, int fdc, off_t start, off_t end) {
	struct _ctb_ds *ds;
	struct _ctb_ds **tmpds;

	/* Create Data source */
	ds = (struct _ctb_ds *)calloc(1, sizeof(struct _ctb_ds));
	if(!ds)
		return -1;

	/* Initialize enough space to hold DS */
	tmpds = (struct _ctb_ds **)realloc(dss->dss,
		(dss->ndss + 1) * sizeof(struct _ctb_ds *));
	if(!tmpds) {
		free(ds);
		return -1;
	}
	dss->dss = tmpds;

	fdc = dup(fdc);
	if(fdc == -1) {
		/* Free the last Data source */
		free(ds);
		return -1;
	}

	ds->off_start = start;
	ds->off_end = end;
	ds->dsfd = fdc;
	dss->dss[dss->ndss] = ds;
	dss->ndss++;

	return 0;
};

void
ctb_free_dss(ctb_dss *dss) {
	int i;

	if(!dss)
		return;

	if(!dss->dss) {
		free(dss);
		return;
	}

	for(i = 0; i < dss->ndss; i++) {
		if(dss->dss[i]) {
			if(dss->dss[i]->dsfd != -1)
				close(dss->dss[i]->dsfd);
			free(dss->dss[i]);
		}
	}

	free(dss->dss);
	free(dss);
};


