#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>

#include "desire.h"
#include "ctbint.h"
#include "session.h"

#ifdef	CTB_RETURN
#undef	CTB_RETURN
#endif

void *
ctb_get(ctb_session *ctbsp, ctb_range *range, int *count) {
	register ctb_sess *ctbs = (ctb_sess *)ctbsp;
	struct _ctb_ds *dsp;
	ctb_dss *dss;
	void *data;
	size_t dataSize;
	ssize_t tmpcnt;
	int cur_ds;

	/* Session has not prepared yet */
	if(!count) return 0;
	if(!ctbs || !range) {
		*count=-1;
		return NULL;
	};

	/* Get data sources */
	dss = ctb_get_dss(ctbsp, range);
	if(!dss) {
		*count = -1;
		return NULL;
	}

	/* Count data size to store all of the data */
	dataSize = ctbs->EntrySize;
	for(cur_ds = 0; cur_ds < dss->ndss; cur_ds++) {
		dsp = dss->dss[cur_ds];
		dataSize += dsp->off_end - dsp->off_start;
	}

	data=(void *)malloc(dataSize);
	if(!data) {
		ctb_free_dss(dss);
		*count = -1;
		return 0;
	}

	*count=0;

	while((tmpcnt = ctb_get_next(ctbsp, dss,
			data + (*count) * ctbs->EntrySize, 1024 * 1024)) > 0)
				*count += tmpcnt;

	ctb_free_dss(dss);
	
	return data;
};


ssize_t
ctb_get_next(ctb_session *ctbsp, ctb_dss *dss, void *datap, size_t size) {
	register ctb_sess *ctbs = (ctb_sess *)ctbsp;
	char *data = (char *)datap;
	struct _ctb_ds *dsp;
	size_t haveBytes = 0;
	int cur_ds;

	if(!ctbs || !dss || !data || size < ctbs->EntrySize)
		return -1;

	/* Necessary alignment */
	if(size % ctbs->EntrySize)
		size -= (size % ctbs->EntrySize);

	/* Iterate through Data sources */
	for(cur_ds = 0; cur_ds < dss->ndss; cur_ds++) {
		ssize_t bytesToGet;
		size_t cycleBytes;

		dsp = dss->dss[cur_ds];

		/* Invalid (closed) file descriptor */
		if(dsp->dsfd == -1)
			continue;

		bytesToGet = (dsp->off_end - dsp->off_start);
		if(bytesToGet <= 0) {
			close(dsp->dsfd);
			dsp->dsfd = -1;
			continue;
		}

		lseek(dsp->dsfd, dsp->off_start, SEEK_SET);
		if(bytesToGet > (size - haveBytes))
			bytesToGet = size - haveBytes;

		/* Read by small portions */
		for(cycleBytes = 0; cycleBytes < bytesToGet;) {
			ssize_t tmpread;

			tmpread = bytesToGet - cycleBytes;
			tmpread = (tmpread > 1024*1024) ? (1024*1024) : tmpread;
			tmpread = read(dsp->dsfd, data + haveBytes + cycleBytes, tmpread);
			if(tmpread <= 0) {
				/* Align */
				syslog(LOG_NOTICE, "tmpread: %d", tmpread);
				bytesToGet = cycleBytes - (cycleBytes % ctbs->EntrySize);
				break;
			}
			cycleBytes += tmpread;
		}

		haveBytes += bytesToGet;
		dsp->off_start += bytesToGet;
		if(dsp->off_start >= dsp->off_end) {
			close(dsp->dsfd);
			dsp->dsfd = -1;
		}

		if(haveBytes >= size)
			break;	/* Return data count */
	}


	return (haveBytes / ctbs->EntrySize);
};


